/* CTCSS generator for ATtiny13
 * Uses external Xtal oscillator @ 1.8432 MHz
 *
 * Pin usage:
 *                +--------------------------------------------------+
 * reset/switch3 -|PB5/PCINT5/RST/ADC0/DW                         VCC|- vcc
 *       Xtal-in -|PB3/PCINT3/CLKI/ADC3        PB2/PCINT2/SCK/ADC1/T0|- switch1
 *       switch2 -|PB4/PCINT4/ADC2     PB1/PCINT1/AIN1/OC0B/INT0/MISO|- ~PTT
 *           gnd -|GND                      PB0/PCINT0/AIN0/OC0A/MOSI|- PWM-out
 *                +--------------------------------------------------+
 *
 * If pin1 is used as RST it reads 1.
 *
 * After 'PTT down' switches are read and choosen frequency is generated
 * When PTT goes up, controller goes to sleep
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>

/* Fdds = FXTAL / 256
 * Usable output freqency ~ 0.4 * Fdds
 *
 *	FXTAL	Fdds	Fmax
 *	1843200	7200	2880
 */
#define	FXTAL 1843200	/* Not used */

// DDS variables
volatile long phase;
volatile long phase_inc;

/* Sine table
 * Minimum value = 0x04
 * Maximum value = 0xfc
 */
const uint8_t  sinewave[] PROGMEM = { //256 values
	0x80,0x83,0x86,0x89,0x8c,0x8f,0x92,0x95,
	0x98,0x9b,0x9e,0xa1,0xa3,0xa6,0xa9,0xac,
	0xaf,0xb2,0xb5,0xb7,0xba,0xbd,0xbf,0xc2,
	0xc4,0xc7,0xc9,0xcc,0xce,0xd0,0xd3,0xd5,
	0xd7,0xd9,0xdb,0xdd,0xdf,0xe1,0xe3,0xe5,
	0xe7,0xe8,0xea,0xeb,0xed,0xee,0xf0,0xf1,
	0xf2,0xf3,0xf4,0xf5,0xf6,0xf7,0xf8,0xf8,
	0xf9,0xfa,0xfa,0xfb,0xfb,0xfb,0xfb,0xfb,
	0xfc,0xfb,0xfb,0xfb,0xfb,0xfb,0xfa,0xfa,
	0xf9,0xf8,0xf8,0xf7,0xf6,0xf5,0xf4,0xf3,
	0xf2,0xf1,0xf0,0xee,0xed,0xeb,0xea,0xe8,
	0xe7,0xe5,0xe3,0xe1,0xdf,0xdd,0xdb,0xd9,
	0xd7,0xd5,0xd3,0xd0,0xce,0xcc,0xc9,0xc7,
	0xc4,0xc2,0xbf,0xbd,0xba,0xb7,0xb5,0xb2,
	0xaf,0xac,0xa9,0xa6,0xa3,0xa1,0x9e,0x9b,
	0x98,0x95,0x92,0x8f,0x8c,0x89,0x86,0x83,
	0x80,0x7d,0x7a,0x77,0x74,0x71,0x6e,0x6b,
	0x68,0x65,0x62,0x5f,0x5d,0x5a,0x57,0x54,
	0x51,0x4e,0x4b,0x49,0x46,0x43,0x41,0x3e,
	0x3c,0x39,0x37,0x34,0x32,0x30,0x2d,0x2b,
	0x29,0x27,0x25,0x23,0x21,0x1f,0x1d,0x1b,
	0x19,0x18,0x16,0x15,0x13,0x12,0x10,0x0f,
	0x0e,0x0d,0x0c,0x0b,0x0a,0x09,0x08,0x08,
	0x07,0x06,0x06,0x05,0x05,0x05,0x05,0x05,
	0x04,0x05,0x05,0x05,0x05,0x05,0x06,0x06,
	0x07,0x08,0x08,0x09,0x0a,0x0b,0x0c,0x0d,
	0x0e,0x0f,0x10,0x12,0x13,0x15,0x16,0x18,
	0x19,0x1b,0x1d,0x1f,0x21,0x23,0x25,0x27,
	0x29,0x2b,0x2d,0x30,0x32,0x34,0x37,0x39,
	0x3c,0x3e,0x41,0x43,0x46,0x49,0x4b,0x4e,
	0x51,0x54,0x57,0x5a,0x5d,0x5f,0x62,0x65,
	0x68,0x6b,0x6e,0x71,0x74,0x77,0x7a,0x7d,
};

/* phase increments
 *
 *             fout * (2^^32)
 * increment = --------------
 *                 fdds
 */
const unsigned long ctcss_inc[] PROGMEM = {
	// Standard config (2 switches), PB5 (RST) reads as '0' (?)
	0x028e7324L /* 71.90 Hz */,
	0x02bcdf01L /* 77.00 Hz */,
	0x02eeeeeeL /* 82.50 Hz */,
	0x03258bf2L /* 88.50 Hz */,
	// Added in extended config (PB5 (RST) is 3rd switch)
	0x0261d950L /* 67.00 Hz */,
	0x08e38e38L /* 250.00 Hz */,
	0x11c71c71L /* 500.00 Hz */,
	0x238e38e3L /* 1000.00 Hz */,
};

/* Called 1.843.200 / 256 = 7200 times per second
 * Maximum frequency approx. 0.4 * 7200 = 2880 Hz
 */
ISR(TIM0_OVF_vect){
	register uint8_t i;

	i = (phase >> 24) & 0xff;
	// arg of pgm_read_byte is an address
	OCR0A = pgm_read_byte(&sinewave[i]);
	phase += phase_inc;
}

// Wakes up from sleep()
EMPTY_INTERRUPT(PCINT0_vect)

int
main(void) {
	volatile register unsigned char phase_index;

	// Set PORTB0 pin as output
	DDRB = _BV(DDB0);

	// PortB0 is OCR0A output
	// COM0A1 - COM0A0 (Set OC0A on Compare Match, clear OC0A at TOP)
	// WGM01 - WGM00 (set fast PWM)
	TCCR0A = (1 << COM0A1) | (1 << COM0A0) | (1 << WGM01) | (1 << WGM00);
	TCCR0B = _BV(CS00); // Start timer at Fcpu / 256
	// initial OCR0A value
	OCR0A = 0x80;

	// Enable pullup on PORTB5, PORTB4, PORTB2, PORTB1
	PORTB = _BV(PB5) | _BV(PB4) | _BV(PB2) | _BV(PB1);

	//enable overflow interrupt for Timer0
	TIMSK0 = _BV(TOIE0);

	// Setup tone parameters
	phase = 0L;

	// Enable pin-changed interupt on PB1
	PCMSK = _BV(PCINT1);
	GIMSK |= _BV(PCIE);

	//enable global interrups
	sei();

	// Main loop
	while (1) {
		/* Read switches
		 * PB2: switch1
		 * PB4: switch2
		 * PB5: switch3
		 * reset reads as 1 if not used for input
		 */
		phase_index = ((PINB >> 2) & 0x01) | ((PINB >> 3) & 0x06);

		// Setup tone parameters
		phase_inc = pgm_read_dword(&ctcss_inc[phase_index]);

		// Wait for 'PTT up'
		if (PINB & _BV(PINB1)) {
			// Disable timer interrupt
			TIMSK0 &= ~_BV(TOIE0);

			// Avoid jumps at start of next cycle
			// Low pass converts this to Vcc/2
			OCR0A = 0x80;

			// Go to sleep
			set_sleep_mode(SLEEP_MODE_PWR_SAVE);
			sleep_mode();

			// Re-enable timer interrupt
			TIMSK0 |= _BV(TOIE0);
		}
	}
}

